/*
 * Copyright 2010 Martin Schreiber
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/*
 * info: OpenGL 3.2 initialization code taken from khronos website
 */

#ifndef CGUI_HPP
#define CGUI_HPP

#include <GL/gl.h>
#include <SDL/SDL.h>
#include <iostream>
#include <sstream>
#include <signal.h>
#include "lib/CError.hpp"
#include "lib/CBitmap.hpp"
#include "lib/CError.hpp"


/**
 * \brief callback handlers for gui events
 */
class CGuiInterface
{
public:
	virtual void callback_key_down(int key, int mod, int scancode, int unicode) = 0;	///< key down event
	virtual void callback_key_up(int key, int mod, int scancode, int unicode) = 0;		///< key up event
	virtual void callback_quit() = 0;							///< quit event
	virtual void callback_mouse_motion(int x, int y) = 0;		///< mouse position in absolute coordinates
	virtual void callback_mouse_button_down(int button) = 0;	///< button press event
	virtual void callback_mouse_button_up(int button) = 0;		///< button release event
	virtual void callback_viewport_changed(int width, int height) = 0;	///< viewport changed (window resized)
};

/**
 * \brief initialize gui and rendering context
 */
class CGui
{
	bool sdl_initialized;

	class CGuiInterface *pcGuiInterface;

	std::string windowTitle;

	class CState
	{
	public:
		Uint32 flags;
		int window_width;
		int window_height;
		float aspect;

		SDL_WindowID windowId;
		SDL_GLContext glContext;
		SDL_DisplayMode mode;
	};

	CState windowState;
	CState fullscreenState;
	CState *runningState;

public:
	CError error;		///< error handler
	CError message;	///< message handler

	bool fullscreen_active;	///< true, if fullscreen is active
	int width;				///< width of viewport/window
	int height;				///< height of viewport/window
	float aspect;			///< current aspect ratio (window/height)

	/**
	 * update viewport with current width and height values and update apsect ratio
	 */
	void updateViewport()
	{
		glViewport(0, 0, width, height);
		aspect = (float)width / (float)height;
	}

	/**
	 * set fullscreen mode or deactivate
	 */
	void setVideoMode(	bool p_fullscreen		///< set to true, if fullscreen should be activated - false otherwise
	)
	{
		if (p_fullscreen)
		{
			SDL_SetWindowDisplayMode(fullscreenState.windowId, &(fullscreenState.mode));

			if (SDL_SetWindowFullscreen(fullscreenState.windowId, 1) < 0)
			{
				SDL_SetWindowDisplayMode(windowState.windowId, &(windowState.mode));
				error << "Failed to initialize fullscreen mode: " << SDL_GetError() << std::endl;
				return;
			}

			SDL_SetWindowSize(fullscreenState.windowId, fullscreenState.window_width, fullscreenState.window_height);

			std::cout << "fullscreen active" << std::endl;
			runningState = &fullscreenState;
			fullscreen_active = true;
		}
		else
		{
//			SDL_SetWindowDisplayMode(windowState.windowId, &(windowState.mode));

			if(SDL_SetWindowFullscreen(windowState.windowId, 0) < 0)
			{
				SDL_SetWindowDisplayMode(fullscreenState.windowId, &(fullscreenState.mode));
				std::cerr << "Failed to initialize window mode: " << SDL_GetError() << std::endl;
				return;
			}

			SDL_ShowWindow(windowState.windowId);
			std::cout << windowState.window_width << windowState.window_height << std::endl;
			SDL_SetWindowSize(windowState.windowId, windowState.window_width, windowState.window_height);

			std::cout << "window mode active" << std::endl;
			runningState = &windowState;
			fullscreen_active = false;
		}
		SDL_ShowWindow(runningState->windowId);

		width = runningState->window_width;
		height = runningState->window_height;
		aspect = (float)width / (float)height;

		updateViewport();

		pcGuiInterface->callback_viewport_changed(runningState->window_width, runningState->window_height);
	}

public:

	/**
	 * initialize GUI
	 */
	CGui(	class CGuiInterface &cGuiInterface,		///< class which implements the gui interface for callback events
			const char *p_windowTitle,				///< title of window to display in the window title bar
			int p_window_width = 800,				///< initial width of window in window mode
			int p_window_height = 600,				///< initial height of window in window mode
			bool p_fullscreen = false					///< set to true to enable fullscreen mode on window creation
	)
	{

		fullscreen_active = p_fullscreen;

		pcGuiInterface = &cGuiInterface;

		windowState.window_width = p_window_width;
		windowState.window_height = p_window_height;

		fullscreenState.window_width = p_window_width;
		fullscreenState.window_height = p_window_height;

		windowTitle = p_windowTitle;

		sdl_initialized = false;
	}

	void create()
	{
		if (SDL_Init(SDL_INIT_VIDEO) < 0) /* Initialize SDL's Video subsystem */
		{
			std::cerr << "Unable to initialize SDL: " << SDL_GetError() << std::endl;
			exit(-1);
		}

//		std::cerr << "a" << SDL_GetError() << std::endl;

		sdl_initialized = true;


		// activate ctrl-c abort handling
		signal(SIGINT, SIG_DFL);
		signal(SIGQUIT, SIG_DFL);
/*
		int major_version;
		if (SDL_GL_GetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, &major_version) == -1)
		{
			error << "Unknown Error requesting GL major version" << std::endl;
			sdl_initialized = false;
			SDL_Quit();
			return;
		}

		if (major_version < 3)
		{
			std::ostringstream errorMessage;
			errorMessage << "OpenGL version > 3.2 necessary. This system has major version " << major_version;
			error << errorMessage.str() << std::endl;
			sdl_initialized = false;
			SDL_Quit();
			return;
		}

		int minor_version;
		if (SDL_GL_GetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, &minor_version) == -1)
		{
			error << "Unknown Error requesting GL minor version" << std::endl;
			sdl_initialized = false;
			SDL_Quit();
			return;
		}

		if (minor_version < 2)
		{
			std::ostringstream errorMessage;
			errorMessage << "OpenGL version > 3.2 necessary. This system has minor version " << minor_version;
			error << errorMessage.str() << std::endl;
			std::cout << "lkjasdf" << std::endl;
			sdl_initialized = false;
			SDL_Quit();
			return;
		}
*/

#if 0
		/* Request an opengl 3.2 context.
		 * SDL doesn't have the ability to choose which profile at this time of writing,
		 * but it should default to the core profile */
		if (SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3) == -1)
		{
			error << "unable to set major OpenGL version to 3" << std::endl;
			sdl_initialized = false;
			SDL_Quit();
			return;
		}

		if (SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2) == -1)
		{
			error << "unable to set minor OpenGL version to 2" << std::endl;
			sdl_initialized = false;
			SDL_Quit();
			return;
		}
#endif
		/* Turn on double buffering with a 24bit Z buffer.
		 * You may need to change this to 16 or 32 for your system */
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
		SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

		// disable vsync
//		SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 1);

		int numDisplayModes = SDL_GetNumDisplayModes();

		// load current mode as window mode
		SDL_GetDesktopDisplayMode(&windowState.mode);
		if (numDisplayModes == 0)
		{
			error << "No Fullscreen mode available" << std::endl;
		}
		else
		{
			message << "Available fullscreen modes:" << std::endl;
			for (int i = 0; i < numDisplayModes; i++)
			{
				SDL_DisplayMode mode;
				SDL_GetDisplayMode(i, &mode);
//				message << " + mode " << i << ": " << mode.w << "x" << mode.h << "@" << mode.refresh_rate << "Hz   refresh rate: " << mode.refresh_rate << std::endl;

				if (i == 0)
				{
					fullscreenState.mode = mode;
					fullscreenState.window_width = mode.w;
					fullscreenState.window_height = mode.h;
				}
			}
		}

		// setup flags
		windowState.flags = SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN | SDL_WINDOW_RESIZABLE;
		fullscreenState.flags = SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN | SDL_WINDOW_FULLSCREEN;

		if (fullscreen_active)		runningState = &fullscreenState;
		else						runningState = &windowState;
/*
		SDL_SetWindowDisplayMode(runningState->windowId, &(runningState->mode));
*/
		windowState.windowId = SDL_CreateWindow(windowTitle.c_str(), SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, runningState->window_width, runningState->window_height, runningState->flags);
		if (!windowState.windowId) /* Die if creation failed */
		{
			error << "Unable to create window: " << SDL_GetError() << std::endl;
			sdl_initialized = false;
			SDL_Quit();
			return;
		}
		fullscreenState.windowId = windowState.windowId;

		if (fullscreen_active)
		{
			if (SDL_SetWindowFullscreen(fullscreenState.windowId, 1) == -1)
			{
				/*
				SDL_SetWindowDisplayMode(windowState.windowId, &(windowState.mode));
				error << "Failed to initialize fullscreen mode: " << SDL_GetError() << std::endl;
				*/
				return;
			}
		}

		/* Create our opengl context and attach it to our window */
		windowState.glContext = SDL_GL_CreateContext(windowState.windowId);
		if (!windowState.glContext)
		{
			error << "Unable to create GL context: " << SDL_GetError() << std::endl;
			sdl_initialized = false;
			SDL_DestroyWindow(runningState->windowId);
			SDL_Quit();
			return;
		}
		fullscreenState.glContext = windowState.glContext;

		width = runningState->window_width;
		height = runningState->window_height;

		SDL_ShowWindow(runningState->windowId);

		/* dont synchronize with vrefresh */
		SDL_GL_SetSwapInterval(0);

		updateViewport();
	}

	~CGui()
	{
		if (sdl_initialized)
		{
			/* Delete our opengl context, destroy our window, and shutdown SDL */
			SDL_GL_DeleteContext(runningState->glContext);
			SDL_DestroyWindow(runningState->windowId);
			SDL_Quit();
		}
	}

	/**
	 * this function has to be called by the rendering loop on every iteration to catch the events and
	 * forward them to the class implementing the callback functions CGuiInterface
	 */
	void eventLoop()
	{
		SDL_Event event;
		// process pending events
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
				case SDL_WINDOWEVENT:
					switch (event.window.event)
					{
						case SDL_WINDOWEVENT_CLOSE:
							pcGuiInterface->callback_quit();
							break;

						case SDL_WINDOWEVENT_RESIZED:
							width = event.window.data1;
							height = event.window.data2;

							runningState->window_width = width;
							runningState->window_height = height;

							updateViewport();
							pcGuiInterface->callback_viewport_changed(width, height);
							break;

					}
					break;

				case SDL_KEYDOWN:
					pcGuiInterface->callback_key_down(event.key.keysym.sym, event.key.keysym.mod, event.key.keysym.scancode, event.key.keysym.unicode);
					break;

				case SDL_KEYUP:
					pcGuiInterface->callback_key_up(event.key.keysym.sym, event.key.keysym.mod, event.key.keysym.scancode, event.key.keysym.unicode);
					break;

				case SDL_QUIT:
					pcGuiInterface->callback_quit();
					break;

				case SDL_MOUSEMOTION:
					pcGuiInterface->callback_mouse_motion(event.motion.x, event.motion.y);
					break;

				case SDL_MOUSEBUTTONDOWN:
					pcGuiInterface->callback_mouse_button_down(event.button.button);
					break;

				case SDL_MOUSEBUTTONUP:
					pcGuiInterface->callback_mouse_button_up(event.button.button);
					break;
			}
		}
	}

	/**
	 * return the ticks as an integer number (DEPRECATED)
	 */
	signed int getTicksInt()
	{
		return SDL_GetTicks();
	}

	/**
	 * return the ticks in seconds
	 */
	double getTicks()
	{
		return (double)SDL_GetTicks()*(1.0/1000.0);
	}

	/**
	 * save a screenshot to a bitmap file
	 * \param filename	filepath to store the screenshot to
	 */
	bool saveScreenshot(std::string &filename)
	{
		CBitmap24 bitmap(width, height);

		glReadPixels(0, 0, width, height, GL_BGR, GL_UNSIGNED_BYTE, bitmap.data);

		if (!bitmap.save(filename))
		{
			error << bitmap.error.getString();
			return false;
		}

		return true;
	}

	/**
	 * convenience handler to call the function with a c-string
	 */
	bool saveScreenshot(const char *filename)
	{
		std::string string_filename(filename);
		return saveScreenshot(string_filename);
	}

	/**
	 * this function has to be called on the end of the rendering loop to swap the front with back buffer
	 */
	void swapBuffer()
	{
		SDL_GL_SwapWindow(runningState->windowId);
	}
};

#endif
